#undef main                 // removes the existing main from the compilers memory

#define G_VERSION           49
#define F_CPU               16000000L     // related to setting clock speed to 16MHz
#define TINY_ARCADE_LED     4             // LED available for blinking on Tiny Arcade pcb at PB4 (physical pin 3)

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>

#include "assets.h"


/*
 
 To do:

 • Get rid of noise in upper right corner of display caused by RAM consumed by adding Melody_Part3_3 to theme_music[]
 • Weird tones sometimes
 • Allow left screen scroll


 Notes:

 digital_write seems to use a lot of memory, iirc, so I've been using these
 to operate the LED for debugging:
 
// turn on led:
// PORTB |= 1 << TINY_ARCADE_LED; //set bit 4

// turn off led:
// PORTB &= ~(1 << TINY_ARCADE_LED); //clear bit 4

 */



/*
================================================================
Game definitions and declarations

  Some of this is fairly constant across games, some changes.
----------------------------------------------------------------
*/

#define    WAITING_TO_START   0
#define    MARIO_DIED         1
#define    RUNNING            2

// The following might not change much for other games:
 
#define    GAME_OVER_SCREEN_DURATION  800
#define    VERSION_SCREEN_DURATION  200

uint8_t    the_game_state = WAITING_TO_START;
uint16_t   g_display_timer;

uint16_t   high_score_eeprom_address = 510; // store high score near top of EEPROM to keep music from trashing it
uint16_t   high_score;
uint8_t    g_reset = 0;
uint32_t   curr_seed = 0;

// The following is fairly specific to the Mario game:
 
#define    MAX_X_FOR_MARIO  72 // 96

enum       mario_state {idle, left, right, squash, dead};
enum       mario_jumpstate {nojump, jumpup, jumpdown};
enum       mario_direction {faceleft, faceright};

#define   NUM_FRAMES_FOR_COIN 3;

uint8_t   coin_frame = 1, 
          coin_timer = 0; 
uint16_t  coin_count = 0;

struct character {
  int16_t   x;
  int16_t   y;
  int16_t   oldx; 
  int16_t   oldy;
  enum      mario_state state;
  enum      mario_jumpstate jumpstate;
  enum      mario_direction dir;
  int8_t    vx;
  int8_t    vy;
  uint8_t   frame;
  uint8_t   collision;
  uint8_t   coincollector;
  uint8_t   mask;
  uint8_t   w;
};

struct character  mario;

void get_world(uint16_t);
void animate_character(struct character *);
void blank_character(uint8_t, struct character *);
void draw_mario(void);
void draw_coin(int, uint8_t);
void handle_map_collisions(struct character *);
void collide_coins(uint8_t, uint8_t, uint8_t);
uint8_t get_coin_y(uint8_t);
uint8_t next_random(uint8_t);


/*
================================================================
Music and sound effects definitions and declarations

  Probably nothing here changes for other games, except maybe TEMPO_DEFAULT
----------------------------------------------------------------
*/

#define SOUNDEFFECTS_SPEAKER 1   // IC pin 6: PB1, Timer 0, OCOB - Sound effects

#define TEMPO_DEFAULT        800

struct music_table 
{
  uint16_t  startpos;
  uint8_t   len;
};

struct melody
{
  uint16_t* melody;         // + 0
  uint8_t*  noteDurations;  // + 2
  uint8_t   numNotes;       // + 3
  uint8_t   progmem;        // + 4
};

struct toneController 
{
  uint8_t   music_index, 
            music_pos, 
            music_note, 
            noteordelay;
  uint16_t  current_delay;
  uint8_t   soundeffect; // 0 = Music, 1 = Sound effect (affects play priority and loop behaviour)
};

struct toneController   MusicController;
struct toneController   SoundEffectController;

uint8_t                *pos;
uint8_t                 buffer_P[60];//[100];
uint8_t                 sound_effect_playing = 0;
uint16_t                MusicSpeed = TEMPO_DEFAULT;
volatile long           tiny_micros = 0;
volatile long           ISR_micro_period = 0; 
volatile long           tone_timer1_toggle_count = 0; 
volatile long           tone_timer0_toggle_count = 0;


void play_sound_effect(uint8_t);
void init_music(struct toneController *);
void handle_music(struct toneController *);
void tiny_tone(unsigned long, unsigned long, uint8_t);


/*
================================================================
Musical score

  This will be different for each game.
----------------------------------------------------------------
*/

#define NUM_ENTRIES   10 // the number of items of type melody in the theme_music[] array (and includes music and sound effects).
#define NUM_ELEMENTS  15 // the number of elements in music_seq
#define COIN_SOUND    NUM_ENTRIES - 2 //3 // this is the index of coin sound in theme_music[]
#define JUMP_SOUND    NUM_ENTRIES - 1 // 4 // this is the index of jump sound in theme_music[]

struct music_table    music_entries[NUM_ENTRIES];
const struct melody   theme_music[]={
                                    {Melody_Intro,   NoteDurations_Intro,   sizeof(Melody_Intro)>>1,   1},  // 0
                                    {Melody_Refrain, NoteDurations_Refrain, sizeof(Melody_Refrain)>>1, 1},  // 1
                                    {Melody_Part2_1, NoteDurations_Part2_1, sizeof(Melody_Part2_1)>>1, 1},  // 2
                                    {Melody_Part2_2, NoteDurations_Part2_2, sizeof(Melody_Part2_2)>>1, 1},  // 3
                                    {Melody_Part2_3, NoteDurations_Part2_3, sizeof(Melody_Part2_3)>>1, 1},  // 4
                                    {Melody_Part3_1, NoteDurations_Part3_1, sizeof(Melody_Part3_1)>>1, 1},  // 5
                                    {Melody_Part3_2, NoteDurations_Part3_2, sizeof(Melody_Part3_2)>>1, 1},  // 6
                                    {Melody_Part3_3, NoteDurations_Part3_3, sizeof(Melody_Part3_3)>>1, 1},  // 7
                                    {CoinSound, CoinSound_Durations, sizeof(CoinSound)>>1, 1},              // 8 - Coin sound effect
                                    {JumpSound, JumpSound_Durations, sizeof(JumpSound)>>1, 1},              // 9 - Jump sound effect
                                    };

// music_seq lists the indices of the elements in theme_music[] in the order in which they're played.
// There are NUM_ELEMENTS elements listed in music_seq: 0 to NUM_ELEMENTS-1. Wrap around on overflow.
const uint8_t         music_seq[] = {0, 1, 1, 2, 3, 2, 4, 2, 3, 2, 4, 5, 6, 5, 7}; 
                                    // for 5 entries: {0, 1, 1, 2, 2, 2, 0, 2, 1, 2, 0, 1};
                                    // for 6 entries: {0, 1, 1, 2, 3, 2, 1, 2, 3, 2, 3, 2};
                                    // for 7 entries: {0, 1, 1, 2, 3, 2, 4, 2, 3, 2, 4, 2};
                                    // for 8 entries: {0, 1, 1, 2, 3, 2, 4, 2, 3, 2, 4, 5};
                                    // for 10 entries: {0, 1, 1, 2, 3, 2, 4, 2, 3, 2, 4, 5, 6, 5, 7};
/*
================================================================
Game video definitions and declarations

  Probably nothing here changes for other games, except maybe WORLD_MAX_BLOCK_COLS
----------------------------------------------------------------
*/
#define   WORLD_MAX_BLOCK_COLS 63  // This is in 8-pixel wide blocks. One screen is 16 blocks wide. So 63 blocks is just shy of four screens wide.

uint8_t   g_screen[WORLD_MAX_BLOCK_COLS + 1];
uint16_t  g_view_x = 0, 
          g_old_view_x = 0, 
          g_r_view_x_trigger;
int       g_delta_view_x = 0;
          
void update_display(void);
void draw_screen(void);
void draw_sprite(int16_t, int16_t, uint8_t, uint8_t, const unsigned char *, uint8_t);
void v_blankout(int, int, uint8_t);


/*
================================================================
OLED definitions and declarations

  Probably nothing here changes for other games
----------------------------------------------------------------
*/
#define   OLED_ADDRESS 0x3C;
#define   OLED_CONTRAST 0xFF //0x7F

#include "oled_lib.c"

void oled_write_integer(uint8_t, uint8_t, uint16_t);


/*
================================================================
Controls definitions and declarations

  Probably nothing here changes for other games unless they don't use three buttons
----------------------------------------------------------------
*/

#define BUTTON_BASE 22

#define BUTTON_LEFT   4     // 4 left 10k
#define BUTTON_JUMP   2     // 5 jump 22k
#define BUTTON_RIGHT  1     // 6 right 47k
#define BUTTONS_ALL   7     // 4+5+6 reset high score 

#define ADC_PIN 3

void      read_buttons(void);
uint16_t  do_ADC(void);



void setup(void);


/*
================================================================
*/



//=============================================================================
//
// Setup proc
//
//=============================================================================


void do_setup() 
{
  pinMode(TINY_ARCADE_LED, OUTPUT);

  //---------------------------------------------------------
  // Graphics setup  

  oledInit(0x3c, 0, 0);
  oledFill(0); // 0 = black screen, 255 = white screen


  //---------------------------------------------------------
  // Game setup  

  mario.mask = 3;
  mario.w = 2;
  
  mario.x = 5;
  mario.y = 0;
  mario.state = idle;
  mario.jumpstate = jumpdown;
  mario.vx = 0;
  mario.vy = 0;
  mario.frame = 0;
  mario.dir = faceright;
  mario.coincollector = 1;

  
  // Set up 2 screens of world for Mario to move through; get_world() is called again later to generate more world.
  get_world(0);  // populates first 16 uint8_t of g_screen[] (indices 0 - 15)
  get_world(16); // populates next 16 uint8_t of g_screen[] (indices 16 - 31)
//  get_world(32); // populates next 16 uint8_t of g_screen[] (indices 32 - 47)
//  get_world(48); // populates last 16 uint8_t of g_screen[] (indices 48 - 63)

  g_r_view_x_trigger = 12; // if g_view_x/8 > trigger get_world element in certain position.

  if (g_reset == 0) 
  {
    g_reset = 1; // Do this stuff once.

    oled_addr = OLED_ADDRESS;

    // Uncomment this and reload this script to set high score to zero,
    // then comment it out and load this again to allow high scores
    // to be recorded. (This all assumes that Tools > Save EEPROM
    // in the Arduino IDE (using TinyCode by Dr. Azzy) is set to
    // "EEPROM retained".) Or just use the "reset_tiny_mario_high_score" script.
//    eeprom_write_word(high_score_eeprom_address, (uint16_t *)0);
    high_score = eeprom_read_word((uint16_t *)high_score_eeprom_address);

    //---------------------------------------------------------
    // Sound setup  

    // Store all music and sound effects in PROGMEM (if melody.progmem = 1, otherwise in EEPROM) only on powerup.
    // This isn't ideal, as there is a limit to the number of writes to EEPROM....
    burn_nonvolatile_memory();

    MusicController.soundeffect = 0;
    SoundEffectController.soundeffect = 1;
    
    sound_effect_playing = 0;
    init_music(&MusicController);
    
    // Set bit 1 of DDRB to 1, to enable sound
    DDRB |= (1 << SOUNDEFFECTS_SPEAKER);
    
    // Prepare other registers for sound
    TCCR0A = 0;
    TIMSK = 0;
    TCCR1 = 0 ;
    GTCCR = 0;
    TIFR = 0;
  }
  else
  {
    MusicController.soundeffect = 0;
    SoundEffectController.soundeffect = 1;
  }
}


//=============================================================================
//
// Main proc
//
//=============================================================================


int main(void)
{
  static long     loop_micros = 0;
  uint16_t        the_block_x;
  uint8_t         the_x, the_y;
  uint8_t         i, k = 0;
  
  init();   // Needed when running from main() rather than standard Arduino usage of setup() and loop().
  do_setup();
  sei();    // Allow interrupts
  
  while (1)
  {
    read_buttons();    

    //------------------------------------------------------
    if (the_game_state == MARIO_DIED)
    {
      if (g_display_timer < GAME_OVER_SCREEN_DURATION) 
      { 
        g_display_timer++;
      }
      else 
      {
        g_display_timer = 0;
        the_game_state = WAITING_TO_START;
        oledFill(0);
      }
    }
    
    //------------------------------------------------------
    if (the_game_state == WAITING_TO_START)
    {
      draw_game_title(30,10);
      draw_high_score(14,30);
      high_score = eeprom_read_word((uint16_t *)high_score_eeprom_address);    
      oled_write_integer(86, 30, high_score, 1);

      if (g_display_timer == 0)
      {
        draw_press_to_start(20,42);            
      }
      else if (g_display_timer == VERSION_SCREEN_DURATION)
      {
        // "Press to start." is displayed at x=20, start erasing from there.
        // Draw 15 spaces to erase prior displayed text, 
        // with each space being CHAR_W in width, so we have 17+15*CHAR_W.
        for (i = 20; i < 20+15*CHAR_W; i += CHAR_W)
        {
          draw_sprite(i, 42, CHAR_W, CHAR_H, &char_SPACE[0], 0);        
        }
        draw_sprite(56, 42, CHAR_W, CHAR_H, &char_v[0], 0);
        oled_write_integer(68, 42, G_VERSION, 0);
      }
      else if (g_display_timer == VERSION_SCREEN_DURATION * 2)
      {
        g_display_timer = -1;        
      }
      g_display_timer++;
    }
    
    //------------------------------------------------------
    if (mario.y == 48)
    {
      // Mario fell in a gap and died.
      // Reset everything.

      mario.state = idle;
      g_view_x = 0;
      g_old_view_x = 0, 
      g_delta_view_x = 0;
      the_block_x = 0; //g_view_x >> 3;
      curr_seed = 0;
      coin_frame = 1; 
      coin_timer = 0;
      do_setup();

      draw_game_over(32,20);

      // From EEPROM docs:
      //    uint16_t eeprom_read_word(const uint16_t *__p)
      //    void     eeprom_write_word(uint16_t *__p, uint16_t __value)
      //
      // read high score from EEPROM
      // if coin_count > high_score, write coin_count to high_score
      high_score = eeprom_read_word((uint16_t *)high_score_eeprom_address);          
      if (coin_count > high_score)
      {
        high_score = coin_count;
        eeprom_write_word(high_score_eeprom_address, high_score);
        draw_new(18,32);        
        draw_high_score(44,32);
        draw_sprite(104, 32, CHAR_W, CHAR_H, &char_BANG[0], 0);
      }
      else
      {
        draw_press_to_start(20,32);
      }
     coin_count = 0;
     the_game_state = MARIO_DIED;
    }

    //------------------------------------------------------
    if (the_game_state == RUNNING)
    {
      handle_map_collisions( &mario ); // Check for collisions with Mario and the world
      
      animate_character(&mario); // just applies gravity, could be done elsewhere (i.e. handle_map_collisions physics?)
      
      // Mario frames go from 0 to 5 inclusive; 0,1,2 get mario_walk1, 3,4,5 get mario_walk2
      if (mario.frame++ >= 5) mario.frame = 0;
      
      coin_timer++;  // Animate coin
      // Coins spin with frames 1, 2, and 3 then repeat 2 and 1, for a total of five frames: 1,2,3,2,1.
      // Hence, 5:
      if (coin_timer >= 5)
      {
        coin_timer = 0;
        coin_frame++;
        if (coin_frame >= 5) coin_frame = 1;
      }

      // Keep the left edge of the scrolling view no more than 86 (or MAX_X_FOR_MARIO if it's not 86) pixels behind Mario. 
      if (mario.x <= MAX_X_FOR_MARIO) g_view_x = 0;
      else if (g_view_x < mario.x - MAX_X_FOR_MARIO) g_view_x = mario.x - MAX_X_FOR_MARIO;

      // g_view_x can be ______ and conceptually refers to pixel coords in x.
      // Convert that to blocks by dividing by 8. (The display is 16 8-pixel blocks wide.)
      the_block_x = g_view_x >> 3; // that is, "= g_view_x / 8"
      if (the_block_x >= g_r_view_x_trigger)
      {
        get_world(g_r_view_x_trigger + 20);
        g_r_view_x_trigger += 8;
      }
      
      draw_coin(0, 0);

      
      // ----------------------------------------------------
      // Probably doesn't change for other games:
      
      update_display(); // Draw screen
      oled_write_integer(10, 0, coin_count, 1);

      handle_music( &MusicController );
      handle_music( &SoundEffectController );
    
      while (tiny_micros - loop_micros <= 25000) // 40 fps
      {
        handle_music( &MusicController );
        handle_music( &SoundEffectController );
      }
      loop_micros = tiny_micros;
    }
    else
    {      
      handle_music( &MusicController );
    }
  }
}


//=============================================================================
//
// Game text procs
//
//   Probably only draw_game_title() needs to be changed for other games.
//=============================================================================

// These procs all display text. Each character is treated as a sprite.

void draw_game_title(uint8_t the_x, uint8_t the_y)
{
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_T[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_i[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_n[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_y[0], 0); the_x += CHAR_W * 2;

  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_M[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_a[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_r[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_i[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_o[0], 0); the_x += CHAR_W;
}


void draw_press_to_start(uint8_t the_x, uint8_t the_y)
{
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_P[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_r[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_e[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_s[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_s[0], 0); the_x += CHAR_W * 2;

  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_t[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_o[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_SPACE[0], 0); the_x += CHAR_W; // erase first digit ovf version number
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_s[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_t[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_a[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_r[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_t[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_PERIOD[0], 0); the_x += CHAR_W;
}


void draw_game_over(uint8_t the_x, uint8_t the_y)
{
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_G[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_a[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_m[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_e[0], 0); the_x += CHAR_W * 2;

  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_O[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_v[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_e[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_r[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_BANG[0], 0); the_x += CHAR_W;
}


void draw_high_score(uint8_t the_x, uint8_t the_y)
{
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_H[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_i[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_g[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_h[0], 0); the_x += CHAR_W * 2;

  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_S[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_c[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_o[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_r[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_e[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_COLON[0], 0); the_x += CHAR_W;
}

void draw_new(uint8_t the_x, uint8_t the_y)
{
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_N[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_e[0], 0); the_x += CHAR_W;
  draw_sprite(the_x, the_y, CHAR_W, CHAR_H, &char_w[0], 0); the_x += CHAR_W;
}


//=============================================================================
//
// Gameplay procs
//
//   A lot of this must be modified for use in games other than Mario.
//=============================================================================


// get_world function is inspired by the "Map::GenerateRoom" function in the Squario project on the Arduboy, found here:
// https://github.com/arduboychris/Squario/blob/master/SquarioGame.cpp

void get_world(uint16_t vx)
{
  // vx is in units of bricks, one screen is 16 bricks wide 
  // (a brick is 8 pixels wide; 16 * 8 px = 128 pixels).
  
  uint8_t the_floor;
  uint8_t the_gap = 0;
  uint8_t cx;
  uint8_t i, j;

  // Note: next_random(n) returns a random number between 0 and n-1, inclusive, up to 255.
 
  cx = (vx & WORLD_MAX_BLOCK_COLS); // this wraps the screen around to vx = 0 when vx = 64, 1 when it's 65, 2 when it's 66, etc.
  the_floor = 5 + next_random(3);  // can be 5, 6 or 7

  // give us the next 16 block-columns of the screen
  for (i = cx; i < cx + 16; i++)
  {
    // Assume there will be no coin, and set first block accordingly (to 0 = no coin).
    g_screen[i] = 0;

    // There's a 1 in 5 chance of spawning a coin (spawned only if next_random() returns 0).
    if (!next_random(5)) g_screen[i] = 1; // Spawn a coin
    
    if (!the_gap)
    {
      // set bits 5,6,7 or 6,7 or 7 of g_screen[this column] if they're floor blocks
      for (j = the_floor; j <= 7; j++)
        g_screen[i] |= (1 << j);

      // Include a gap if next_random() returns 0 (so a 1 in 9 chance)
      // AND if we're not in the first column.
      if (!next_random(9)  &&  cx != 0) //vx != 0)
      {
        // set the gap width to at least 3, and up to 5.
        the_gap = 3 + next_random(3); 
      }
      else if (!next_random(5) && cx != 0) //vx != 0) // 1 in 5 chance of changing the floor height if not in the first column
      {
        // 1 in 2 chance of increasing floor height by 1 as far as page 7 (whihc I guess is actually DOWN), otherwise decrease it by 1 as far as to page 1
        if (!next_random(2) && the_floor < 7) the_floor++;
        else if (the_floor > 1) the_floor--;
      }
    } 
    else the_gap--;
   }
}


void animate_character(struct character *player)
{
  // Jumping
  if (player->jumpstate != nojump)
  {
    player->vy += 1; // Gravity
    if (player->vy >= 0 && player->jumpstate == jumpup) player->jumpstate = jumpdown; // flip state for animation
  }
}


void blank_character(struct character *the_player)
{
  // Blankout old mario position sprite if we've moved
  if ((the_player->x - g_view_x) != the_player->oldx || (the_player->y != the_player->oldy)) 
  {
    v_blankout(the_player->oldx, the_player->oldy, 16);
    v_blankout(the_player->oldx, the_player->oldy + 8, 16);
  }

  the_player->oldx = the_player->x - g_view_x;
  the_player->oldy = the_player->y;
}


void draw_mario()
{
  if (mario.jumpstate == nojump) // so, idle, left, or right
  {
    if (mario.state == idle) draw_sprite(mario.x - g_view_x, mario.y, 16, 16, &mario_idle[0], mario.dir);
    else if (mario.state == left || mario.state == right)
    {
      // Mario frames go from 0 to 5 inclusive; 0,1,2 use mario_walk1, 3,4,5 use mario_walk2
      if (mario.frame < 3) draw_sprite(mario.x - g_view_x, mario.y, 16, 16, &mario_walk1[0], mario.dir);
      else draw_sprite(mario.x - g_view_x, mario.y, 16, 16, &mario_walk2[0], mario.dir);
    } // end: left or right walking
  } // end: not jumping
  else if (mario.jumpstate == jumpup) draw_sprite(mario.x - g_view_x, mario.y, 16, 16, &mario_jump1[0], mario.dir);
  else if (mario.jumpstate == jumpdown) draw_sprite(mario.x - g_view_x, mario.y, 16, 16, &mario_jump2[0], mario.dir);
}


void draw_coin(int the_sprite_x, uint8_t the_sprite_y)
{
  if (g_delta_view_x < 0) v_blankout(the_sprite_x + 8, the_sprite_y, -g_delta_view_x);
  else if (g_delta_view_x > 0) v_blankout(the_sprite_x - g_delta_view_x, the_sprite_y, g_delta_view_x);

  if (coin_frame == 1)
    draw_sprite(the_sprite_x, the_sprite_y, 8, 8, &coin1[0], 0);
  else if (coin_frame == 2 || coin_frame == 4)
    draw_sprite(the_sprite_x, the_sprite_y, 8, 8, &coin2[0], 0);
  else if (coin_frame == 3)
    draw_sprite(the_sprite_x, the_sprite_y, 8, 8, &coin3[0], 0);
}


void handle_map_collisions(struct character *player)
{
  // Only apply velocities if no collisions in that direction, else clip to collision point.
  int       newmario_x, newmario_y;
  uint8_t   cell_x, cell_y, ybitmask;
  uint16_t  cell_x_zone;
  uint8_t   check_cell, check_cell2;
  uint8_t   the_x_offset, the_y_offset;

  if (player->vy < -7) player->vy = -7;   
  newmario_x = player-> x + player->vx;
  newmario_y =  player->y + player->vy;
  player->collision = 0; // 1 = UP, 2=RIGHT, 4=DOWN, 8=LEFT

  cell_x_zone = (newmario_x >> 9); // 0-31 = 0, 32-63=1, 64-95=2, 96-127=3. >>8 = 3 + (log2 WORLD_BUFFER);

  // 0-63 = 0, 64-127 = 1 ... therefore 64*8 = 2^6 * 2^3 = 2^9 = 512
  cell_x = (newmario_x >> 3) & WORLD_MAX_BLOCK_COLS;
  cell_y = newmario_y >> 3; // generate bit-mask for byte describing y-column in screen data.
  the_y_offset = newmario_y & 0x07;
  the_x_offset = newmario_x & 0x07;

  //ybitmask = 1 << cell_y;
  if (the_y_offset == 0) ybitmask = (player->mask) << cell_y;
  else ybitmask = ((player->mask<<1)+1)<<cell_y; // add an extra cell 
  
  if (newmario_y < 0) 
  {
    cell_y = 0;  // means that a block is never found and mario can jump through the top of the screen!
    ybitmask = 0;
  }
  else if (newmario_y>48) // can't fall off the bottom of the screen!
  {
    newmario_y = 48;
    player->vy = 0;    
    cell_y = 6;  
    ybitmask = 0;
  }

  ybitmask&=~(1<<0); // means no collisions with row0

  if (player->vx > 0) // Check Mario's righthand side if he tried to move to the right
  {                 
    check_cell = (cell_x + player->w) & WORLD_MAX_BLOCK_COLS; // WORLD_BUFFER - 1
    
    //if (screen[check_cell] & ybitmask || screen[check_cell] & (ybitmask << 1) || (the_y_offset != 0 && screen[check_cell] & (ybitmask << 2)))
    if (g_screen[check_cell] & ybitmask) // that's all
    { // clip
      // Strictly we should update cell_x_zone to be relevant to the zone of the block that caused the collision
      player->x = (cell_x_zone << 9) + (cell_x << 3);
      if (cell_x == 0) cell_x = WORLD_MAX_BLOCK_COLS; 
      else cell_x -= 1; // handle wrap
      player->collision |= 2; // 1 = UP, 2=RIGHT, 4=DOWN, 8=LEFT
      player->vx = 0;
    }
    else
    {
      player->x = newmario_x;
    }
  }
  else if (player->vx < 0)
  {
    //player->x=newmario_x;
   // if (g_screen[cell_x]&ybitmask || g_screen[cell_x] & (ybitmask << 1) || (the_y_offset != 0 && g_screen[cell_x] & (ybitmask << 2)))
   if (g_screen[cell_x] & ybitmask) 
    { // clip
      ///cell_x=(cell_x+1) & 31; // if collided with left, push back to the right - handle wrap
      cell_x++;
      if (cell_x > WORLD_MAX_BLOCK_COLS) 
      {
        cell_x = 0;
        cell_x_zone++;
      }
      player->x = (cell_x_zone << 9) + (cell_x << 3); // Strictly we should update cell_x_zone to be relevant to the zone of the block that caused the collision
      player->collision |= 8; // 1 = UP, 2=RIGHT, 4=DOWN, 8=LEFT
      player->vx = 0;
    }
    else
    {
      player->x = newmario_x;
    }
  }

  //
  // cell_y+2 should be valid at this point
  
  ybitmask = 1 << (cell_y + player->w);
  ybitmask &= ~(1 << 0); // means no collisions with row 0
  
  if (player->vy > 0 || player->vy == 0) // going down
  {
    check_cell = (cell_x + 1) & WORLD_MAX_BLOCK_COLS;
    check_cell2 = (cell_x + 2) & WORLD_MAX_BLOCK_COLS;
    //if ((cell_y + 2 >= 8) || g_screen[cell_x] & ybitmask  || g_screen[check_cell] & ybitmask  || (the_x_offset != 0 && g_screen[check_cell2] & ybitmask))
    if (g_screen[cell_x] & ybitmask  || g_screen[check_cell] & ybitmask  || (the_x_offset != 0 && g_screen[check_cell2] & ybitmask))
    { // clip
      player->vy = 0; player->jumpstate = nojump;
      if (cell_y + 2 >= 8) cell_y = 6;
      player->y = (cell_y << 3);
      player->collision |= 4; // 1 = UP, 2=RIGHT, 4=DOWN, 8=LEFT
      player->vy = 0;
    }
    else
    {
      if (player->vy == 0 && player->jumpstate == nojump) {
        player->jumpstate = jumpdown;
      }
      player->y = newmario_y; // allow new vertical position
    }
  }
  else if (player->vy < 0) // Have we hit our head?
  {
    ybitmask = 1 << (cell_y);
    ybitmask&=~(1<<0); // means no collisions with row 0
  
    //player->y=newmario_y;// allow new vertical position
    check_cell = (cell_x + 1) & WORLD_MAX_BLOCK_COLS; // WORLD_BUFFER -1
    check_cell2 = (cell_x + 2) & WORLD_MAX_BLOCK_COLS; // WORLD_BUFFER -1

    if (g_screen[cell_x] & (ybitmask) || g_screen[check_cell] & (ybitmask) || (the_x_offset != 0 && g_screen[check_cell2] & (ybitmask)))
    { // clip
      player->vy = 0; player->jumpstate = jumpdown;
      cell_y += 1;
      player->y = (cell_y << 3);
//      if (player->coincollector) play_sound_effect(HITHEAD_SOUND);
      player->collision |= 1; // 1 = UP, 2=RIGHT, 4=DOWN, 8=LEFT
      player->vy = 0;
    }
    else
    {
      //if (mario.vy==0 && mario.jumpstate==nojump) {mario.jumpstate=jumpdown;}
      player->y = newmario_y; // allow new vertical position
    }
  }
  // Coin collision

  // mario is at: cell_x, cell_x+1 (and if the_x_offset!=0, cell_x+2)
  // and at: cell_y, cell_y+1 (and if the_y_offset!=0, cell_y+2)

  if (player->coincollector)
  {
    collide_coins(cell_x, cell_y, the_y_offset);
    collide_coins(cell_x + 1, cell_y, the_y_offset);
    if (the_x_offset != 0) collide_coins(cell_x + 2, cell_y, the_y_offset);
  }
}


void collide_coins(uint8_t the_block_col, uint8_t the_page_y, uint8_t the_y_offset)
{
  uint8_t the_coin_y = 8; 
  uint8_t h;

  h = 2; 
  if (the_y_offset != 0) h = 3;
  the_block_col = the_block_col & WORLD_MAX_BLOCK_COLS;
  if (g_screen[the_block_col] & 1) // Is there a coin in this column?
  {
    the_coin_y = get_coin_y(the_block_col);
    if (the_coin_y != 8)
    {
      if (the_coin_y >= the_page_y && the_coin_y < (the_page_y + h))
      {
        g_screen[the_block_col] &= 254;
        play_sound_effect(COIN_SOUND);
        coin_count++;
        // Reset coin count if it's about to get bigger than uint16_t.
        if (coin_count == 65535) 
        {
          coin_count = 0;
          high_score = 0;
          eeprom_write_word(high_score_eeprom_address, high_score);
        }
      }
    }
  }
}


uint8_t get_coin_y(uint8_t the_block_col)
{
  uint8_t y;

  for (y = 1; y < 8; y++)
  {
    // Locate the highest block in this column.
    // if ... AND bit y of g_screen[n] is a 1
    if (g_screen[the_block_col & WORLD_MAX_BLOCK_COLS] & (1 << y)) // There is a block here
    {
      if (y == 1) return 0; //draw_coin(the_x_offset, (y-1)<<3);
      else return y - 2; //draw_coin(the_x_offset, (y-2)<<3);
    }
  }

  // 8 means couldn't find a coin. 
  // This would be an error because we tested for a coin before this proc was called.
  return 8; 
}


uint8_t next_random(uint8_t range)
{
  uint8_t the_return_val;
  
  // Returns a random 8-bit integer between 0 and range inclusive (up to 255).
  curr_seed++;
  curr_seed = (curr_seed*1664525 + 1013904223) % 65536;
  return(uint8_t) (curr_seed % range);
}


//=============================================================================
//
// Game-specific graphics procs
//
//   This must be modified more or less, depending on the game.
//=============================================================================


void update_display()
{
  g_delta_view_x = g_old_view_x - g_view_x; // < 0 means moving right, > 0 means moving left.
  if (g_delta_view_x > 8) g_delta_view_x = 8; // Limit g_delta_view_x
  if (g_delta_view_x < -8) g_delta_view_x = -8;
  draw_screen(); // Draws level scenery, coins, etc.
   
  // Draw sprites etc
  blank_character(&mario);
  draw_mario();
  g_old_view_x = g_view_x; // left screen
}


/*
  On the OLED, we can't draw to individual pixels, only to 1 pixel wide x 8 pixel tall groups.
  In each group, the least significant bit (bit 0) is the top bit and the most significant bit 
  (bit 7) is the bottom one.
  
  The display is 128w x 64h, so we have 8 rows of these groups (each row is called a page), 
  and 128 columns of them.
  
  If instead of 1-pixel wide columns we use 8-pixel wide columns, then with 8-pixel rows,
  we have 8x8 blocks, and the entire screen can be represented with 16 columns (of 8x8 blocks),
  each column having 8 rows; so an 8x16 array of blocks. 
  
  Each column can be encoded as a 16-bit number (i.e. as in int), where 0 in a bit means
  no graphic in that block, and 1 means there is a graphic there. The actual graphic that's
  there is determined from context, except for the top row of blocks; these indicate that a
  coin should be drawn in the second block above the topmost floor block (the top row of blocks 
  never has anything draw in them, so they're free for encoding coin locations as described).
  This encoding into ints, like drawing to the oled, has LSBs at the top, and MSBs at the bottom.

  As the screen image scrolls and as Mario moves, only areas that had a graphic block (but don't 
  anymore) get erased if necessary, rather than redrawing the entire screen. This makes screen 
  updates faster.
*/


void draw_screen()
{
  uint8_t the_starting_block_column, 
          the_ending_block_column, 
          wrapped_i = 0, 
          the_coin = 0;
  int     the_x_offset;
  uint8_t the_block_col, 
          the_page;

  // WORLD_MAX_BLOCK_COLS = 63, which in binary is 0b111111
  // Note: 
  //  g_view_x >> 3 = g_view_x/8
  //  the_page << 3 = the_page * 8

  // Get the col corresponding to the left edge of the screen content (as given by g_view_x), 
  // and the screen content's offset from the left edge of that column, if any, wrapping around 
  // at col 64 (or whatever WORLD_MAX_BLOCK_COLS + 1 is).
  // the_x_offset can be 0 through -7.
  the_starting_block_column = (g_view_x >> 3) & WORLD_MAX_BLOCK_COLS;
  the_x_offset = -(g_view_x & 0x07);
  
  // Handle blanking blacking out edge when *just* disappeared off screen. (?)
  if (the_starting_block_column > 0) 
  {
    the_starting_block_column--;  
    the_x_offset -= 8;
  }

  // If there's an offset, we need to add an extra brick on the right because we are not drawing left brick fully,
  // otherwise the ending col is the one that gives us one screen full from the starting col (i.e. 16 cols).
  if (the_x_offset != 0) the_ending_block_column = the_starting_block_column + 16 + 1; 
  else the_ending_block_column--;

  // loop through each column of 8x8 blocks from left to right
  for (the_block_col = the_starting_block_column; the_block_col < the_ending_block_column; the_block_col++) // limit of 32 for two oleds, 16 for one oled
  {
    if (g_delta_view_x < 0) wrapped_i = (the_block_col + 1) & WORLD_MAX_BLOCK_COLS; // wrap round (could be WORLD_BUFFER - 1)
    else if (g_delta_view_x > 0) wrapped_i = (the_block_col - 1) & WORLD_MAX_BLOCK_COLS;

    the_coin = 0;
    // if there's a 1 in bit 0 of the_block_col, then there's a coin in some page of this column.
    if (g_screen[the_block_col & WORLD_MAX_BLOCK_COLS] & (1 << 0)) the_coin = 1;

    // loop through each page (row of 8x8 blocks) in the current column from top to bottom
    for (the_page = 1; the_page < 8; the_page++)
    {
      // if there's a 1 in the bit for this page in the_block_col, 
      // then there's a block in this column of this page.
      if (g_screen[the_block_col & WORLD_MAX_BLOCK_COLS] & (1 << the_page)) 
      {
        if (g_delta_view_x < 0) // screen is scrolling right and we are within array limit
        {
          // therefore blank out smearing trail to the right of block
          // but only if there isn't a block there.
          if ( !(g_screen[wrapped_i] & (1 << the_page)) ) 
          {
            v_blankout(the_x_offset + 8, the_page << 3, -g_delta_view_x); // View moving to the right, blocks scrolling to the left,          
          }
        }
 
        if (g_delta_view_x > 0)
        {
          if (!(g_screen[wrapped_i] & (1 << the_page))) v_blankout(the_x_offset - g_delta_view_x, the_page << 3, g_delta_view_x);          
        }

        // if there's NOT a 1 in the bit for the page above in this column (i.e. there's nothing to draw above it),
        // then draw a topbrick sprite, otherwise draw a regular brick.
        if ((g_screen[the_block_col & WORLD_MAX_BLOCK_COLS] & (1 << (the_page - 1))) == 0)
        {
          draw_sprite(the_x_offset, the_page << 3, 8, 8, &topbrick[0], 0);
        }
        else draw_sprite(the_x_offset, the_page << 3, 8, 8, &brick8x8[0], 0);

        if (the_coin)
        {
          // if this column has a coin, and the floor is near the top of the screen, put coin just above floor
          // otherwise float coin in second block above floor.
          if (the_page == 1) draw_coin(the_x_offset, (the_page - 1) << 3);
          else draw_coin(the_x_offset, (the_page - 2) << 3);
          the_coin = 0;
        }
      }
    }
    the_x_offset += 8;
  }
}


//=============================================================================
//
// Controls procs
//
//  Some of this will need to be modified for other games.
//=============================================================================

/* 
  Convert sensor value to digits 1 through 7, where 1,2,3 are from single resistors,
  4,5,6 are from pairs of resistors, and 7 is from all three resistors.
  This works because we're doing division on ints, so the fractional parts of the 
  results are discarded.
  
  Example: Button 3 has a 47K resistor. With the other resistor in the circuit (1K),
  pressing button 3 gives a sensor value of 21. (That comes from (1024/(47K+1K))*1K, 
  truncated to an integer value.) Then,
    the_buttons = (the_sensor_value + (BUTTON_BASE / 2)) / BUTTON_BASE
    = (21 + (22/2)) / 22
    = (21 + 11) / 22
    = (32) / 22
    = 1.47 BUT this is integer division, so we get just 1, which is
    what BUTTON_RIGHT is defined as at the top of this file.
  
  This program doesn't check for simultaneous button presses, but pressing 
  the_buttons 1 (10K) and 2 (22K) would give a sensor value of 130. 
  (That comes from R = (10K*22K)/(10K+22K) = 6875, then (1024/(R+1K))*1K, 
  again truncated.) Then,
    the_buttons = (the_sensor_value + (BUTTON_BASE / 2)) / BUTTON_BASE
    = (130 + (22/2)) / 22
    = (130 + 11) / 22
    = (142) / 22
    = 6.45 BUT again this is integer division, so we get just 6, which is
    what BUTTONS_1_AND_2 would be defined as at the top of this file if we were
    using that comnbination. (But we're not, so it's not defined up there.)
  
  Herewith, the complete list of sensor values:
  Button(s)____Resistor(s)_____Sensor value_____Result
  1            10K             93               4
  2            22K             45               2
  3            47K             21               1
  1+2          10K+22K         130              6
  1+3          10K+47K         111              5
  2+3          22K+47K         64               3
  1+2+3        10K+22K+47K     146              7
*/

void read_buttons(void)
{
  uint16_t  the_sensor_value;
  uint8_t   the_buttons;
  int       i;

  the_sensor_value = do_ADC();
  the_buttons = (the_sensor_value + (BUTTON_BASE / 2)) / BUTTON_BASE;

  // -------------------------------------------------
  // Start game
  
  if (the_buttons && the_game_state != RUNNING) 
  { 
    the_game_state = RUNNING; 
    oledFill(0); 
    return; 
  }

  // -------------------------------------------------
  // Lateral movement

  // "&& mario.x >= 5 && mario.x > g_view_x" in the else, below,
  // prevents Mario from moving off screen left.
  // It would be better to fix the left dcrolling, but we'll
  // save that work for another time....
  if (the_buttons & BUTTON_RIGHT)
  {
    mario.state = right; 
    mario.dir = faceright;
    mario.vx = 3;
  }
  else if (the_buttons & BUTTON_LEFT && mario.x >= 5 && mario.x > g_view_x)
  {
    mario.state = left; 
    mario.dir = faceleft;
    mario.vx = -3;
  }
  else
  {
    // If button not pressed but still moving (vx > 0), decelerate
    if (mario.vx > 0) mario.vx -= 1;
    else if (mario.vx < 0) mario.vx += 1;
  }


  // -------------------------------------------------
  // Jumping
  // Only start jumping if not currently jumping, and even then only 
  // if we're not facing left near the left side of the screen.
  
  if ((the_buttons & BUTTON_JUMP) && mario.jumpstate == nojump) 
  {
    {
      mario.vy = -7;
      mario.jumpstate = jumpup;
      play_sound_effect(JUMP_SOUND);
    }
  }
  
  // -------------------------------------------------
  // Idle 
  // If not jumping and no keys pressed
  if (the_buttons == 0 && mario.jumpstate == nojump) 
  {
    mario.state = idle;
  }

  //-------------------------------------------------------
  // This is for testing the voltages through the switches.
  //
/*  
  if (0)
  {
    the_sensor_value = 0;
    for (i = 0; i < 8; i++)
    {
      the_sensor_value += do_ADC();
    }
    the_sensor_value = the_sensor_value >> 3;
  
    oled_write_integer(40, 10, the_sensor_value, 1);
  }
*/
  //
  //-------------------------------------------------------
}





//=============================================================================
//=============================================================================
//=============================================================================
//
//
//  Probably nothing below here needs to be changed.
//
//
//=============================================================================
//=============================================================================
//=============================================================================



//=============================================================================
//
// Analog-Digital-Conversion for contol buttons.
//
//  Probably nothing here changes for other games.
//=============================================================================

/*
   -------------------------------------------
   ADMUX - ADC Multiplexer Selection Register
   -------------------------------------------
   Bit - Name  - Read/Write
    7  - REFS1 - R/W
    6  - REFS0 - R/W
    5  - ADLAR - R/W
    4  - REFS2 - R/W
    3  - MUX3  - R/W
    2  - MUX2  - R/W
    1  - MUX1  - R/W
    0  - MUX0  - R/W

  Voltage Reference Selection Bits
  REFS2 - REFS1 - REFS0 - Vref Selection
    X   -   0   -   0   - VCC used as Voltage Reference, disconnected from PB0 (AREF).
    X   -   0   -   1   - External Voltage Reference at PB0 (AREF) pin, Internal Voltage Reference turned off.
    0   -   1   -   0   - Internal 1.1V Voltage Reference.
    0   -   1   -   1   - Reserved
    1   -   1   -   0   - Internal 2.56V Voltage Reference without external bypass capacitor, disconnected from PB0 (AREF).
    1   -   1   -   1   - Internal 2.56V Voltage Reference with external bypass capacitor at PB0 (AREF) pin.

   -------------------------------------------
   ADCSRA – ADC Control and Status Register A
   -------------------------------------------
   Bit - Name  - Read/Write
    7  - ADEN  - R/W
    6  - ADSC  - R/W
    5  - ADATE - R/W
    4  - ADIF  - R/W
    3  - ADIE  - R/W
    2  - ADSP2 - R/W
    1  - ADSP1 - R/W
    0  - ADSP0 - R/W

   ADC Prescaler Select Bits
   ADPS2 - ADPS1 - ADPS0 - Division factor
     0   -   0   -   0   -   2
     0   -   0   -   1   -   2
     0   -   1   -   0   -   4
     0   -   1   -   1   -   8
     1   -   0   -   0   -   16
     1   -   0   -   1   -   32
     1   -   1   -   0   -   64
     1   -   1   -   1   -   128

   Input Channel Selections
   MUX3 - MUX2 - MUX1 - MUX0 - Input
    0   -  0   -  0   -  0   - Single ended input ADC0 (PB5)
    0   -  0   -  0   -  1   - Single ended input ADC1 (PB2)
    0   -  0   -  1   -  0   - Single ended input ADC2 (PB4)
    0   -  0   -  1   -  1   - Single ended input ADC3 (PB3)
    ... table continues

*/

/* 
   Buttons 1,2,3 go to PB5 (the Tiny Arcade pcb is set up to enable this, but this program does not make use of ADC0 on pin PB5)
   Buttons 4,5,6 go to PB3 (this is the ADC Tiny Mario uses).
*/

uint16_t do_ADC()
{
  uint16_t  value = 0;
  uint8_t   i;

  // Repeat 16 times so that we can get mean value.
  for (i = 0; i < 16; i++)
  {
    // ADEN is bit 7 of ADCSRA; 
    // 1 << ADEN gives 1000 0000 so ~(1 << ADEN) gives 0111 1111; 
    // ANDing it with whatever is in ADCSRA sets bit 7 to 0, disabling ADC.
    ADCSRA &= ~(1 << ADEN); 

    // ADMUX's REFSx bits are all 0 when initialized, and we don't change that, so the Vref is Vcc
    // (see table above).
    
    // ADC pin is 3 (0011 binary)
    // So ADC_PIN gives 0011, which sets MUX0 and MUX1 to 1, 
    // thus selecting ADC3 (PB3) for ADC input, and single ended conversion.
    ADMUX = ADC_PIN;
    
    // Set ADEN (ADCSRA bit 7) to 1, enabling ADC; 
    // and set the 3 ADPSx bits to 1 as well to set the division factor between the system clock frequency 
    // and the input clock to the ADC to 16 (see table above).
    ADCSRA = (1 << ADEN) | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0);

    // ADSC is bit 6 of ADCSRA;
    // 1 << ADSC gives 0100 0000;
    // ORing it with whatever is in ADCSRA sets  bit 6 to 1, triggering start of ADC conversion.
    ADCSRA |= (1 << ADSC);

    // ADSC will read as one as long as a conversion is in progress. 
    // When the conversion is complete, it returns to zero. 
    // Writing zero to this bit has no effect.
    // So we wait until ADSC goes to 0, at which time the conversion is complete and we can move on.
    // Again, 1 << ADSC gives 0100 0000;
    // so while bit 6 of ADCSRA is 1, loop and do nothing.
    while( ADCSRA & (1 << ADSC));

    // In single ended conversion, the result of the conversion is given by
    //   ADC = (Vin * 1024)/Vref
    // And the results of the conversion are stored in the ADc data register.
    value += ADC;
  }
  
  // Divide by 16 to get mean value.
  return value >> 4; 
}


//=============================================================================
//
// Generic graphics procs
//
//   Probably nothing here changes for other games.
//=============================================================================


void oled_write_integer(uint8_t x, uint8_t y, uint16_t number, uint8_t pad_flag)
{
  uint16_t  n;
  uint16_t  n_div = 10000;
  uint8_t   i;
  uint8_t   the_digit_written_flag = 0;
  
  oledSetPosition(x, y);

  for (i = 0; i < 5; i++)
  {
    n = number / n_div; // 0-255, so n is never > 10
    // Write the number n if it's not zero, or, if it IS zero, only if pad_flag is 1 or it's not the first digit
    if (n != 0 ||  (n == 0 && pad_flag == 1)  ||  (n == 0 && the_digit_written_flag == 1) )
    {
      the_digit_written_flag = 1;
      oledWriteDataBlock(&digitsFont[n * 6], 6); // write character pattern
    }
    number = number - n_div * n;
    n_div = n_div / 10; // 10000 -> 1000 -> 100 --> 10 --> 1
  }
}

void draw_sprite(int16_t x, int16_t y, uint8_t w, uint8_t h, const unsigned char *buf, uint8_t flip)
{
  int16_t        the_sprite_x, the_sprite_y;
  unsigned char *pos = (unsigned char*)buf;
  uint8_t        start_height = 0, 
                 stop_height = h/8, 
                 the_x_offset = 0, 
                 the_visible_width = w;
  uint8_t        temp[16];
  uint8_t        i, j;
  
  // write_height 16 means 2 lots of 8 bits.

  if (x >= 128 || x <= -w) return;
  if (y > 63) return;
  
  if (x < 0) 
  {
    the_x_offset = -x;
    x = 0;
  }
  if (x + w > 128) the_visible_width = 128 - x;

  
  if (y < 0)
  {
    start_height = 1 + ((-y) >> 3); // i.e. -7 to -1 (inclusive): start_height = 1
    y = 0;
  }
  
  if (y + h > 64) stop_height = 8 - (y >> 3);
  
  for (j = start_height; j < stop_height; j++)
  {

    // Read into temp[], the sprite data pointed to by *buf.
    // Note that h >> 3 = h/8.
    for (i = the_x_offset; i < the_visible_width; i++)
    {
      if (flip == 0) pos = (unsigned char*)buf + j + (i * (h >> 3));
      else pos = (unsigned char*)buf + j + ((w - 1 - i) * (h >> 3));
      temp[i - the_x_offset] =  pgm_read_byte(pos);
    }

    // Position cursor before drawing.
    the_sprite_x = x;
    the_sprite_y = y + (j - start_height) * 8;
    oledSetPosition(the_sprite_x, the_sprite_y);

    // Draw the data in temp[] to the screen.
    oledWriteDataBlock(temp, the_visible_width - the_x_offset);  
  }
}


void v_blankout(int x, int y, uint8_t the_width)
{
  uint8_t the_blank_data[16] = {0};

  // Adjust the x and y location if necessary, and calculate the width
  // of blank data to send to oled. Then send it.

  // If area to blank is off screen left or right, do nothing and return.
  if (x + the_width <= 0 || x >= 128) return;

  // If area to blank is off screen top or bottom, do nothing and return.
  if (y < -7 || y > 63) return;

  // If area to blank is partially off screen left or right,
  // subtract from the width that portion of width that's off screen.
  // Otherwise if the area to blank extends off the right side of the screen,
  // then set the width to just what's on screen.
  if (x < 0) 
  {
    the_width = x + the_width;
    x = 0;
  }
  else if (x + the_width > 128) 
  {
    the_width = 128 - x;
  }
 
  if (y < 0) y = 0; 
  
  oledSetPosition(x, y);
  oledWriteDataBlock(the_blank_data, the_width);
}


//=============================================================================
//
// Stuff for burning music and sound effects to EEPROM
//
//   Probably nothing here changes for other games.
//=============================================================================


void burn_nonvolatile_memory(void)
{
  uint8_t i;
  
  pos=(NUM_ENTRIES * sizeof(music_table)); // Leave room for entry table at start
  for (i = 0; i < NUM_ENTRIES; i++)
    burn_music(i); 
}


uint8_t* burn_music(uint8_t entry)
{
  uint8_t   i;
  uint16_t *melody;
  uint8_t  *noteDurations;
  uint8_t   progmem, 
            len;
  uint8_t   duration,
            power;
  
  // Extract the good stuff
  melody = theme_music[entry].melody;
  noteDurations = theme_music[entry].noteDurations;
  progmem = theme_music[entry].progmem;
  len = theme_music[entry].numNotes;
  
  // Sort the table entry out
  music_entries[entry].len = len;
  music_entries[entry].startpos = pos;
  eeprom_write_block( &music_entries[entry], entry*sizeof(music_table), sizeof(music_table) );
  
  // Write notes in next pos
  if (progmem)
  {
    memcpy_P(&buffer_P[0], melody, len * 2);
  }
  else
  {
    memcpy(&buffer_P[0], melody, len * 2);
  }

  // melody in buffer_P
  for (i = 0; i < len; i++)
  {
    if (progmem) duration = pgm_read_byte(noteDurations + i);
    else duration = noteDurations[i];
    
    power = 1;
    if (duration == 2) power = 1;
    else if (duration == 4) power=2;
    else if (duration == 8) power=3;
    else if (duration == 16) power=4;
    else if (duration == 32) power=5;
    else if (duration == 64) power=6;
    else if (duration == 128) power=7;
    
    buffer_P[1+(i*2)] |= (power<<5); // bytes are swapped in uint16_t
  }
  eeprom_write_block(buffer_P, pos, len*2); // 2 bytes per note  
  pos += (len*2);
}


//=============================================================================
//
// Sound procs
//
//  Probably nothing here changes for other games.
//=============================================================================


void init_music(struct toneController *controller)
{
  controller->music_pos = 0; // First position in music sequence, i.e. 0,1,2,3 etc.
  controller->music_index = music_seq[controller->music_pos];
  controller->music_note = 0; // First note of music_index melody
  controller->noteordelay = 1; //simulate end of a delay to start the note
}


void play_sound_effect(uint8_t num)
{
  if (sound_effect_playing) return; // only one at once

  tone_timer0_toggle_count = 0; // stop current sound

  SoundEffectController.music_pos = 0;
  SoundEffectController.music_index = num;
  SoundEffectController.noteordelay = 1;

  sound_effect_playing = 1;
}


void handle_music(struct toneController *controller)
{
  uint16_t *melody;
  uint8_t   numNotes;
  uint8_t  *entry_pos;
  uint16_t  ms;
  uint16_t  freq;
  uint8_t   note;
  

  // This was here before... huh?:
  // soundeffect = 1: Timer0 is used for music - SPEAKER2
  // soundeffect = 0: Timer1 is used for sound effects - SPEAKER 

  if ((
        controller->soundeffect == 0 && 
        tone_timer1_toggle_count == 0
      ) 
   || ( controller->soundeffect == 1 && 
        tone_timer0_toggle_count == 0 && 
        sound_effect_playing == 1
      )
     )
  {
   if (controller->noteordelay == 0) // 0 = NOTE just ended, therefore now delay
    {
      // Note ended, start a delay
      controller->current_delay = controller->current_delay / 3;

      if (controller->soundeffect == 0) tiny_tone(0, controller->current_delay, 1);
      else tiny_tone(0, controller->current_delay, 0);

      // Advance to next note position
      controller->music_note++;               // sizeof(music_table)
      numNotes = eeprom_read_byte((uint8_t *)(controller->music_index * 3) + 2); // next position is numNotes

      if (controller->music_note >= numNotes) // exceeded notes in this melody part
      {
        controller->music_note = 0;
        controller->music_pos++; // advance sequence position

        if (controller->soundeffect == 1) // we are in the sound effects handler
        {
          sound_effect_playing = 0; // Global; control is passed back to music handler; music will restart at next tone.
        }
        else if (controller->music_pos >= NUM_ELEMENTS) // In music handler, and sequence has overflowed.
        {
          controller->music_pos = 1; // We are in the music handler
        }
        controller->music_index = music_seq[controller->music_pos];
      }
      controller->noteordelay = 1; //delay is next
    } // end note ended
    else if (controller->noteordelay == 1) // 1 = DELAY
    {
      // Delay ended, play next note
      entry_pos = (uint8_t *)(controller->music_index * 3);//sizeof(music_table);
      melody = (uint16_t *)eeprom_read_word((uint16_t*)entry_pos); // EEPROM position in entry table
      entry_pos += 2;

      ms = 1000;

      if (controller->soundeffect == 0) ms = MusicSpeed; // Music speed might change

      freq = eeprom_read_word(melody + controller->music_note);  // duration is now encoded into top 3 bits of frequency.
      note = freq >> 13; // Extract 2,4,8,16th,32nd note etc from freq
      freq &= 0x1FFF;
      controller->current_delay = ms>>note; // shift rather than divide.
    
      if (controller->soundeffect == 0) tiny_tone(freq, controller->current_delay, 1);
      else tiny_tone(freq, controller->current_delay, 0); // Sound effects
     
      controller->noteordelay = 0;
    }// end delay ended
  }
}

void tiny_tone(unsigned long frequency, unsigned long duration, uint8_t timer)
{
  uint32_t  ocr;
  uint8_t   prescalar_bits = 0b001;
  uint8_t   output = 1;

  // This is hard coded for pin PB1 (OCR0A).
  
  if (frequency == 0)
  { 
    // Means we want to switch off the pin and do a pause, but not output to the hardware pin
    output = 0; 
    frequency = 300; // A dummy frequency - timings are calculated but no pin output.
  }

  ocr = F_CPU / (2 * frequency);
  
  if (timer == 0)
  {
    if (ocr > 256)
    {
      ocr >>= 3; //divide by 8
      prescalar_bits = 0b010;  // ck/8
      if (ocr > 256)
      {
        ocr >>= 3; //divide by a further 8
        prescalar_bits = 0b011; //ck/64
        if (ocr > 256)
        {
          ocr >>= 2; //divide by a further 4
          prescalar_bits = 0b100; //ck/256
          if (ocr > 256)
          {
            // can't do any better than /1024
            ocr >>= 2; //divide by a further 4
            prescalar_bits = 0b101; //ck/1024
          } // ck/1024
        }// ck/256
      }// ck/64
    }//ck/8*/
  } //timer == 0
  else
  {
    prescalar_bits = 1;
    while (ocr > 0xff && prescalar_bits < 15) 
    {
      prescalar_bits++;
      ocr>>=1;
    }
  }
   
  ocr -= 1;

  if (timer == 0)
  {
    tone_timer0_toggle_count = (2 * frequency * duration) / 1000;
  }
  else
  {
    // timer 1 scalar code here
    tone_timer1_toggle_count = (2 * frequency * duration) / 1000;
    // Music uses timer1 and is on continuously so use as a crude timer
    ISR_micro_period = 500000L / frequency; // 1E6 / (2*frequency), because ISR is called 2 times every period.
  }

  /*
   -------------------------------------------
   TCCROA - Timer/Counter Control Register A
   -------------------------------------------
   Bit - Name   - Read/Write
    7  - COM0A1 - R/W
    6  - COM0A0 - R/W
    5  - COM0B1 - R/W
    4  - COM0B0 - R/W
    3  - --     - R
    2  - --     - R
    1  - WGM01  - R/W
    0  - WGM00  - R/W

  COM0A1 - COM0A0
  COM0B1 - COM0B0 - Description when in CTC or Normal mode (see below).
    0    -   0    - Normal port operation; OC0A/OC0B disconnected.
    0    -   1    - Toggle OC0A/OC0B on Compare Match.
    1    -   0    - Clear OC0A/OC0B on Compare Match.
    1    -   1    - Set OC0A/OC0B on Compare Match.
  
  Mode - WGM02 - WGM01 - WGM00 - Timer/Counter Mode - TOP  - Update OCRx at - TOV Flag Set on
   0   -   0   -   0   -   0   - Normal             - 0xFF - Immediate      - MAX
   1   -   0   -   0   -   1   - PWM, Phase Correct - 0xFF - TOP            - BOTTOM
   2   -   0   -   1   -   0   - CTC                - OCRA - Immediate      - MAX
   3   -   0   -   1   -   1   - Fast PWM           - 0xFF - BOTTOM         - MAX
   4   -   1   -   0   -   0   - Reserved           - --   - --             - --
   5   -   1   -   0   -   1   - PWM, Phase Correct - OCRA - TOP            - BOTTOM
   6   -   1   -   1   -   0   - Reserved           - --   - --             - --
   7   -   1   -   1   -   1   - Fast PWM           - OCRA - BOTTOM         - TOP
 
   -------------------------------------------
   TCCROB - Timer/Counter Control Register B
   -------------------------------------------
   Bit - Name   - Read/Write
    7  - FOC0A  - W  
    6  - FOC0B  - W  
    5  - --     - R  
    4  - --     - R  
    3  - WGM02  - R/W  
    2  - CS02   - R/W  
    1  - CS01   - R/W  
    0  - CS00   - R/W

    WGM02 in register B is used with WGM0x bits in register A to set timer 0's mode (see table above).
    CS00, CS01, and CS02 are used to prescale the input clock frtequency:

    CS02 - CS01 - CS00
     0   -  0   -  0   - No clock source; timer/counter stopped.
     0   -  0   -  1   - clk / 1 (i.e. no prescaling)    
     0   -  1   -  0   - clk / 8     
     0   -  1   -  1   - clk / 64     
     1   -  0   -  0   - clk / 256     
     1   -  0   -  1   - clk / 1024     
     1   -  1   -  0   - External clock source on pin T0; clock on falling edge.    
     1   -  1   -  1   - External clock source on pin T0; clock on rising edge.    
      
   */
  if (output)
  {
    if (timer == 0)
    {
      // Timer/Counter Control Register A
      // Set TCCR0A register to 0001 0010 (where 1 << COM0B0 gives 0001 0000 and 1 << WGM01 gives 0000 0010; OR them to get 0001 0010).
      // So toggle OC0A/OC0B and set mode to CTC.
      TCCR0A = (1 << COM0B0) | (1 << WGM01); 
    }
  }
  else // no output 
  {
    if (timer == 0)
    {
      // (1 << COM0B0) = 0001 0000, so ~(1 << COM0B0) gives 1110 1111; ANDing that with whatever is in TCCR0A effectively sets bit 4 to 0.
      // So no hardware pin output.
      TCCR0A &= ~(1 << COM0B0); 
      // As above, 1 << WGM01 gives 0000 0010; ORing it with whatever is in TCCR0A sets bit 1 to 1.
      // Set mode to CTC.
      TCCR0A |= (1 << WGM01);   
    }
    else
    {
      // timer 1 code to turn off hardware support here
       TCCR1 &= (1 << COM1A0);  // disconnect OC1A
    }
  }

  if (timer == 0)
  {
     // Timer/Counter Control Register B
    TCCR0B = prescalar_bits << CS00;
    TCNT0 = 0; // Initialize counter 0 value to 0
    OCR0A = ocr; // Set compare value
    TIMSK |= (1 << OCIE0A); // Activate timer 0 compare interrupt (timer0 COMPA ISR)
  }
  else
  {
    // Timer/Counter Control Register 1
    TCCR1 = (1 << CTC1) | (prescalar_bits << CS10); // CTC1 : Clear Timer/Counter on Compare Match, after compare match with OCR1C value
    if (output) TCCR1 |= (1 << COM1A0);
    TCNT1 = 0; // Initialize counter 1 value to 0
    OCR1C = ocr; // Set compare value 
    TIMSK |= (1 << OCIE1A); // Activate timer 1 compare interrupt (timer1 COMPA ISR)
  }
}

// Timer 0 interrupt request
ISR(TIMER0_COMPA_vect) 
{
  if (tone_timer0_toggle_count > 0)
  {
    tone_timer0_toggle_count--;
    if (tone_timer0_toggle_count == 0)
    {
      // turn off tone
      // (1 << COM0B0) = 0001 0000, so ~(1 << COM0B0) gives 1110 1111; ANDing that with whatever is in TCCR0A effectively sets bit 4 to 0.
      TCCR0A &= ~(1 << COM0B0);
      // (1 << OCIE0A) = 0001 0000, so ~(1 << OCIE0A) gives 1110 1111; ANDing that with whatever is in TIMSK effectively sets bit 4 to 0.
      TIMSK &= ~(1 << OCIE0A); // Timer/Counter0 Output Compare Match A Interrupt Enable... is now disabled.
      return;
    }
  }
}

// Timer 1 interrupt request
ISR(TIMER1_COMPA_vect) 
{
  if (tone_timer1_toggle_count > 0)
  {
    tone_timer1_toggle_count--;
    if (tone_timer1_toggle_count == 0)
    {
      // turn off tone
      TCCR1 &= (1 << COM1A0); 
      return;
    }
  }
  tiny_micros += ISR_micro_period;
}
